
"""
Installs whisper in an isolated environment.
"""

import os
import shutil
from pathlib import Path

from isolated_environment import IsolatedEnvironment  # type: ignore

TENSOR_VERSION = "2.1.2"
CUDA_VERSION = "cu121"
TENSOR_CUDA_VERSION = f"{TENSOR_VERSION}+{CUDA_VERSION}"
EXTRA_INDEX_URL = f"https://download.pytorch.org/whl/{CUDA_VERSION}"

HERE = Path(os.path.abspath(os.path.dirname(__file__)))

def has_nvidia_smi() -> bool:
    """Returns True if nvidia-smi is installed."""
    return shutil.which("nvidia-smi") is not None


def unit_test() -> None:
    """Unit test."""
    from tempfile import TemporaryDirectory  # pylint: disable=import-outside-toplevel
    with TemporaryDirectory() as tmpdir:
        print(f"Using temporary directory {tmpdir}")
        iso_env = IsolatedEnvironment(HERE / 'transcribe_anything_env')
        iso_env.install_environment()
        iso_env.pip_install('torch==2.1.2', EXTRA_INDEX_URL)
        iso_env.pip_install('openai-whisper')
        iso_env.run(['whisper', '--help'])

if __name__ == "__main__":
    unit_test()



# pylint: skip-file
# flake8: noqa

"""
Runs whisper api.
"""

import shutil
import sys
import time
import json
from pathlib import Path
import subprocess
from typing import Optional, Any
from filelock import FileLock

from isolated_environment import IsolatedEnvironment  # type: ignore
from transcribe_anything.cuda_available import CudaInfo

HERE = Path(__file__).parent
ENV: Optional[IsolatedEnvironment] = None
CUDA_INFO: Optional[CudaInfo] = None
ENV_LOCK = FileLock(HERE / "insane_whisper_env.lock")

# Set the versions
TENSOR_VERSION = "2.1.2"
CUDA_VERSION = "cu121"
TENSOR_CUDA_VERSION = f"{TENSOR_VERSION}+{CUDA_VERSION}"
EXTRA_INDEX_URL = f"https://download.pytorch.org/whl/{CUDA_VERSION}"

def get_current_python_version() -> str:
    """Returns the current python version."""
    return f"{sys.version_info.major}.{sys.version_info.minor}.{sys.version_info.micro}"

# for insanely fast whisper, use:
#   pipx install insanely-fast-whisper --python python3.11

def has_nvidia_smi() -> bool:
    """Returns True if nvidia-smi is installed."""
    return shutil.which("nvidia-smi") is not None

def get_environment() -> IsolatedEnvironment:
    """Returns the environment."""
    global ENV  # pylint: disable=global-statement
    with ENV_LOCK:
        if ENV is not None:
            return ENV
        venv_dir = HERE / "venv" / "insanely_fast_whisper"
        env = IsolatedEnvironment(venv_dir)
        if not venv_dir.exists():
            env.install_environment()
            if has_nvidia_smi():
                env.pip_install(f"torch=={TENSOR_VERSION}", extra_index=EXTRA_INDEX_URL)
            else:
                env.pip_install(f"torch=={TENSOR_VERSION}")
            env.pip_install("openai-whisper")
            env.pip_install("insanely-fast-whisper")
        ENV = env
        return env


def get_cuda_info() -> CudaInfo:
    """Get the computing device."""
    global CUDA_INFO  # pylint: disable=global-statement
    if CUDA_INFO is None:
        iso_env = get_environment()
        env = iso_env.environment()
        py_file = HERE / "cuda_available.py"
        cp: subprocess.CompletedProcess = subprocess.run([
            "python", py_file
        ], check=False, env=env, universal_newlines=True, stdout=subprocess.PIPE)
        stdout = cp.stdout
        CUDA_INFO = CudaInfo.from_json_str(stdout)
    return CUDA_INFO

def get_device_id() -> str:
    """Get the device id."""
    # on mac, we just return "mps"
    if sys.platform == "darwin":
        return "mps"
    cuda_info = get_cuda_info()
    if not cuda_info.cuda_available:
        raise ValueError("CUDA is not available.")
    device_id = cuda_info.cuda_devices[0].device_id
    return f"{device_id}"

def get_batch_size() -> int | None:
    """Returns the batch size."""
    if sys.platform == "darwin":
        return 4
    return None

def convert_time_to_srt_format(timestamp: float) -> str:
    """Converts timestamp in seconds to SRT time format (hours:minutes:seconds,milliseconds)."""
    hours, remainder = divmod(timestamp, 3600)
    minutes, seconds = divmod(remainder, 60)
    milliseconds = int((seconds % 1) * 1000)
    seconds = int(seconds)
    return f"{hours:02}:{minutes:02}:{seconds:02},{milliseconds:03}"

def convert_json_to_srt(json_data: dict[str, Any]) -> str:
    """Converts JSON data from speech-to-text tool to SRT format."""
    srt_content = ""
    for index, chunk in enumerate(json_data['chunks'], start=1):
        start_time, end_time = chunk['timestamp']
        start_time_str = convert_time_to_srt_format(start_time)
        end_time_str = convert_time_to_srt_format(end_time)
        text = str(chunk['text']).strip()
        srt_content += f"{index}\n{start_time_str} --> {end_time_str}\n{text}\n\n"
    return srt_content

def convert_json_to_text(json_data: dict[str, Any]) -> str:
    """Converts JSON data from speech-to-text tool to text."""
    text = ""
    for chunk in json_data['chunks']:
        text += str(chunk['text']).strip() + "\n"
    return text


def run_insanely_fast_whisper(  # pylint: disable=too-many-arguments
    input_wav: Path,
    model: str,
    output_dir: Path,
    task: str,
    language: str,
    other_args: Optional[list[str]]
) -> None:
    """Runs insanely fast whisper."""
    iso_env = get_environment()
    device_id = get_device_id()
    cmd_list = []
    output_dir.mkdir(parents=True, exist_ok=True)
    outfile = output_dir / "out.json"
    model = f"openai/whisper-{model}"
    if sys.platform == "win32":
        # Set the text mode to UTF-8 on Windows.
        cmd_list.extend(["chcp", "65001", "&&"])
    cmd_list += [
        "insanely-fast-whisper",
        "--file-name", str(input_wav),
        "--device-id", f"{device_id}",
        "--model-name", model,
        "--task", task,
        "--transcript-path", str(outfile),
    ]
    if language:
        cmd_list += ["--language", language]
    batch_size = get_batch_size()
    if batch_size is not None:
        cmd_list += ["--batch-size", f"{batch_size}"]
    if other_args:
        cmd_list.extend(other_args)
    # Remove the empty strings.
    cmd_list = [x.strip() for x in cmd_list if x.strip()]
    cmd = " ".join(cmd_list)
    sys.stderr.write(f"Running:\n  {cmd}\n")
    proc = subprocess.Popen(  # pylint: disable=consider-using-with
        cmd, shell=True, universal_newlines=True,
        encoding="utf-8", env=iso_env.environment()
    )
    while True:
        rtn = proc.poll()
        if rtn is None:
            time.sleep(0.25)
            continue
        if rtn != 0:
            msg = f"Failed to execute {cmd}\n "
            raise OSError(msg)
        break
    assert outfile.exists(), f"Expected {outfile} to exist."
    json_text = outfile.read_text(encoding="utf-8")
    json_data = json.loads(json_text)
    srt_content = convert_json_to_srt(json_data)
    srt_file = output_dir / "out.srt"
    txt_content = convert_json_to_text(json_data)
    srt_file.write_text(srt_content, encoding="utf-8")
    txt_file = output_dir / "out.txt"
    txt_file.write_text(txt_content, encoding="utf-8")





# pylint: skip-file

"""
Queries the system for CUDA devices and returns a json string with the information.
This is meant to be run under a "isolated-environment".
"""

import shutil
import sys
from dataclasses import dataclass, asdict, fields
import json
from typing import Any


@dataclass
class CudaDevice:
    """Data class to hold CUDA device information."""
    name: str
    vram: int  # VRAM in bytes
    multiprocessors: int  # Number of multiprocessors
    device_id: int

    def __str__(self):
        return (f"{self.name} - VRAM: {self.vram / (1024 ** 3):.2f} GB, "
                f"Multiprocessors: {self.multiprocessors}")
    def to_json(self) -> dict[str, str | int]:
        """Returns a dictionary representation of the object."""
        return asdict(self)

    @staticmethod
    def from_json(json_data: dict[str, str | int]) -> 'CudaDevice':
        """Returns a CudaDevice object from a dictionary."""
        return CudaDevice(**json_data)  # type: ignore

@dataclass
class CudaInfo:
    """Cuda info."""
    cuda_available: bool
    num_cuda_devices: int
    cuda_devices: list[CudaDevice]

    def to_json_str(self) -> str:
        """Returns json str."""
        # Convert dataclass to dictionary for serialization
        data = self.to_json()
        return json.dumps(data, indent=4, sort_keys=True)

    def to_json(self) -> dict[str, Any]:
        """Returns a dictionary representation of the object."""
        out = {}
        for field in fields(self):
            out[field.name] = getattr(self, field.name)
            if field.name == 'cuda_devices':
                out[field.name] = [device.to_json() for device in out[field.name]]
        return out

    @staticmethod
    def from_json_str(json_str: str) -> 'CudaInfo':
        """Loads from json str and returns a CudaInfo object."""
        data = json.loads(json_str)
        cuda_devices_data = data.get('cuda_devices', [])
        cuda_devices = [CudaDevice(**device) for device in cuda_devices_data]
        return CudaInfo(data['cuda_available'], data['num_cuda_devices'], cuda_devices)

def cuda_cards_available() -> CudaInfo:
    """
    Returns a CudaInfo object with information about the CUDA cards,
    ordered by VRAM and multiprocessors.
    """
    # Have to import here, since others will import CudaDevice and CudaInfo.
    if shutil.which('nvidia-smi') is None:
        return CudaInfo(False, 0, [])
    import torch  # type: ignore
    if torch.cuda.is_available():
        devices = [
            CudaDevice(
                name=torch.cuda.get_device_name(i),
                vram=torch.cuda.get_device_properties(i).total_memory,
                multiprocessors=torch.cuda.get_device_properties(i).multi_processor_count,
                device_id=i
            ) for i in range(torch.cuda.device_count())
        ]
        # Sort devices by VRAM and then by number of multiprocessors in descending order
        devices.sort(key=lambda x: (x.vram, x.multiprocessors), reverse=True)
        return CudaInfo(True, len(devices), devices)
    return CudaInfo(False, 0, [])

def main() -> int:
    """Returns 0 if cuda is available, 1 otherwise."""
    cuda_info = cuda_cards_available()
    print(cuda_info.to_json_str())
    if cuda_info.cuda_available:
        return 0
    return 1

if __name__ == '__main__':
    sys.exit(main())







"""
Tests transcribe_anything
"""

# pylint: disable=bad-option-value,useless-option-value,no-self-use,protected-access,R0801
# flake8: noqa E501

import unittest
import json

from transcribe_anything.insanely_fast_whisper import convert_json_to_srt, convert_json_to_text

EXAMPLE_JSON = """
{
  "speakers": [],
  "chunks": [
    {
      "timestamp": [
        0,
        3.24
      ],
      "text": " Oh wow, I'm so nervous."
    },
    {
      "timestamp": [
        3.24,
        6.56
      ],
      "text": " Gosh, these lights are so bright."
    },
    {
      "timestamp": [
        6.56,
        8.52
      ],
      "text": " Is this mic on?"
    },
    {
      "timestamp": [
        8.52,
        9.52
      ],
      "text": " Is there even a mic?"
    }
  ],
  "text": " Oh wow, I'm so nervous. Gosh, these lights are so bright. Is this mic on? Is there even a mic?"
}
"""

class JsonToSrtTester(unittest.TestCase):
    """Tester for transcribe anything."""

    def test_json_to_srt(self) -> None:
        """Check that the command works on a local file."""
        data = json.loads(EXAMPLE_JSON)
        out = convert_json_to_srt(data)
        print(out)
        print()

    def test_json_to_txt(self) -> None:
        """Check that the command works on a local file."""
        data = json.loads(EXAMPLE_JSON)
        out = convert_json_to_text(data)
        print(out)
        print()


if __name__ == "__main__":
    unittest.main()
